#ifndef __POOL_FOR_STRINGS_TERMINATED_BY_ZERO_H_
#define __POOL_FOR_STRINGS_TERMINATED_BY_ZERO_H_

// This library helps keep many short strings.
// StringPool makes a copy of an added string and returns its address.
// Then, that copy is deleted when the class is deleted.
//
// StringPool uses large blocks to store strings so that the overhead of
// new [] and delete [] becomes little. The default size of each block is
// specified by the argument of its construcor.
//
// When a string which is longer than the default block size is passed,
// StringPool allocates a large block to that string.

// For assert().
#include <cassert>
// For std::size_t and NULL.
#include <cstdio>

// For vector.
#include <vector>


namespace StringPool {

using std::size_t;

// This class is to store many short strings.
template <typename Letter>
class Pool {
 private:
  typedef std::vector<Letter *> BlockVector;
  typedef typename BlockVector::iterator BlockIterator;

 public:
  Pool(size_t block_size)
      : blocks_(), next_(NULL), end_(NULL),
        num_strs_(0), total_size_(0), block_size_(block_size) {}
  ~Pool() {
    for (BlockIterator it = blocks_.begin(); it != blocks_.end(); ++it)
      delete [] *it;
  }

  // Add a copy of a string to the pool.
  Letter *add(const Letter *s, size_t size) {
    assert(s != NULL);
    if (size > static_cast<size_t>(end_ - next_))
      expand_(size);
    strcpy_(next_, s, size);
    Letter *start = next_;
    next_ += size;
    ++num_strs_;
    return start;
  }

  Letter *add(const Letter *s) {
    assert(s != NULL);
    return add(s, strlen_(s) + 1);
  }

  // Return the number of strings and the total size.
  size_t num_strs() const { return num_strs_; }
  size_t total_size() const { return total_size_; }

 private:
  BlockVector blocks_;
  Letter *next_;
  Letter *end_;
  size_t num_strs_;
  size_t total_size_;
  size_t block_size_;

  // Allocate memory to a new block.
  void expand_(size_t size) {
    size = (size > block_size_) ? size : block_size_;
    blocks_.push_back(new Letter[size]);
    next_ = blocks_.back();
    end_ = next_ + size;
    total_size_ += sizeof(Letter) * size;
  }

  // Copy a string.
  void strcpy_(Letter *dst, const Letter *src, size_t len) {
    for (size_t i = 0; i < len; ++i)
      dst[i] = src[i];
  }

  // Calculate the length of a string.
  size_t strlen_(const Letter *s) {
    const Letter *p = s;
    while (*p != '\0')
      ++p;
    return static_cast<size_t>(p - s);
  }

  // Prohibit copy and assignment.
  Pool(const Pool &rhs);
  Pool &operator=(const Pool<Letter> &rhs);
};

}  // namespace StringPool.


#endif  // __POOL_FOR_STRINGS_TERMINATED_BY_ZERO_H_
